.equ TV_LATCH,LATB
.equ TV_TRIS,TRISB
.equ TV_PORT,PORTB
.equ TV_SYNC,6
.equ TV_PICTURE,5

.equ TV_TIMER_MS_PERIOD,40

; Latches used for outputting data and control signals.
.equ LCD_CONTROL_LATCH, LATB

; Bit indices within the control latch corresponding to LCD control pins.
.equ LCD_FLM, 15
.equ LCD_M, 14
.equ LCD_C1, 13
.equ LCD_C2, 12
.equ LCD_D_OFF, 11

.text

.global _vdc_tv_retrace_init
_vdc_tv_retrace_init:
	push w0
	mov #handle(long_sync_sync),w0
	mov w0,_vdc_retrace_function_ptr
	mov #5,w0
	mov w0,_vdc_tv_retrace_function_repeat_count
	pop w0
	return

.global _vdc_lcd_retrace_init
_vdc_lcd_retrace_init:
	push w0
	mov #handle(lcd_retrace),w0
	mov w0,_vdc_retrace_function_ptr
	pop w0
	return

.global __T1Interrupt
__T1Interrupt:

	; Preserve registers used by this function on the stack.
	push w0
	push w1

	bclr IEC0,#3 ; T1IE

	mov _vdc_retrace_function_ptr,w0
	goto w0

lcd_retrace: ; Retrace the LCD.

	push w2
	push w3
	
	; Are we starting a new frame?
	cp0 _vdc_active_line_counter
	bra nz, not_starting_new_frame			
		; Toggle the M pin to provide the LCD AC voltage.
		btg LCD_CONTROL_LATCH, #LCD_M
		nop
		; Reset the scanning pointer.
		mov #edsoffset(_vdc_frame_buffer), w0
		mov w0,_vdc_frame_scan_ptr
	not_starting_new_frame:

	; Begin the line.
	bset LCD_CONTROL_LATCH, #LCD_C1
	nop
	bclr LCD_CONTROL_LATCH, #LCD_C1
	nop

	; Clear the frame start line...
	bclr LCD_CONTROL_LATCH, #LCD_FLM
	nop
	bra nz, keep_flm_low
		; ...unless we are actually starting a new frame.
		bset LCD_CONTROL_LATCH, #LCD_FLM
		nop
	keep_flm_low:


	mov #80,w3

lcd_column_loop:
	
	; Output a column.
	
	mov _vdc_frame_scan_ptr, w0

	btsc w3, #1
	bra _output_column_1

	_output_column_0:
	
		btsc w3,#0
		bra _output_column_0_1
	
	_output_column_0_0: ; W3:00 -> 1st part
		mov [w0],w1
		lsr w1,#12,w1
		bra _output_column_data

	_output_column_0_1: ; W3:01 -> 4th part
		mov [w0++],w1
		mov w0,_vdc_frame_scan_ptr
		bra _output_column_data
	
	_output_column_1:
		btsc w3,#0
		bra _output_column_1_1
	
	_output_column_1_0: ; W3:10 -> 3rd part
		mov [w0],w1
		lsr w1,#4,w1
		bra _output_column_data

	_output_column_1_1: ; W3:11 -> 2nd part
		mov [w0],w1
		lsr w1,#8,w1

_output_column_data:
	; Output the data and issue the clock signal.

	mov #4,w0
	btst.z w1,#2
	bsw.z w1,w0

	btsc w1,#3
	bset LATB,#0
	nop
	
	btss w1,#3
	bclr LATB,#0
	nop

	bset LCD_CONTROL_LATCH, #LCD_C2
	mov w1, LATA
	bclr LCD_CONTROL_LATCH, #LCD_C2
	nop

	dec w3,w3
	bra nz,lcd_column_loop

	; Increment the row counter.
	inc _vdc_active_line_counter

	; Have we reached line 240?
	mov #240, w0
	cp _vdc_active_line_counter
	bra nz, not_finished_frame
		; Yes, so reset the row counter.
		clr _vdc_active_line_counter
	not_finished_frame:
	
	pop w3
	pop w2
	pop w1
	pop w0

	; Clear the interrupt flag.
	bclr IFS0,#3 ; T1IF
	bset IEC0,#3 ; T1IE
	retfie

long_sync_sync: ; Output the sync portion of a long sync.
	bclr TV_LATCH,#TV_SYNC
	mov #handle(long_sync_black),w0
	mov #(30 * TV_TIMER_MS_PERIOD - 1),w1
	goto exit_retrace

long_sync_black: ; Output the black portion of a long sync.
	bset TV_LATCH,#TV_SYNC
	mov #(2 * TV_TIMER_MS_PERIOD - 1),w1
	dec _vdc_tv_retrace_function_repeat_count
	bra z,long_sync_to_short_sync
	; Repeat long sync.
	mov #handle(long_sync_sync),w0
	goto exit_retrace
long_sync_to_short_sync:
	; Switch to short sync.
	mov #5,w0
	mov w0,_vdc_tv_retrace_function_repeat_count
	mov #handle(short_sync_pre_frame_sync),w0
	goto exit_retrace

short_sync_pre_frame_sync: ; Output the sync portion of a short sync.
	bclr TV_LATCH,#TV_SYNC
	mov #handle(short_sync_pre_frame_black),w0
	mov #(2 * TV_TIMER_MS_PERIOD - 1),w1
	goto exit_retrace

short_sync_pre_frame_black: ; Output the sync portion of a short sync.
	bset TV_LATCH,#TV_SYNC
	mov #(30 * TV_TIMER_MS_PERIOD - 1),w1
	dec _vdc_tv_retrace_function_repeat_count
	bra z,short_sync_pre_frame_to_picture
	; Repeat short sync.
	mov #handle(short_sync_pre_frame_sync),w0
	goto exit_retrace
short_sync_pre_frame_to_picture:
	; Switch to picture.
	mov #304,w0
	mov w0,_vdc_tv_retrace_function_repeat_count
	
	mov #edsoffset(_vdc_frame_buffer),w0
	mov w0,_vdc_frame_scan_ptr
	
	neg _vdc_tv_line_offset_y,wreg
	mov wreg,_vdc_active_line_counter
	
	mov #handle(picture_sync),w0
	goto exit_retrace

picture_sync: ; Output the sync pulse of a picture line.
	bclr TV_LATCH,#TV_SYNC
	mov #handle(picture_back_porch),w0
	mov #(4 * TV_TIMER_MS_PERIOD - 1),w1
	goto exit_retrace

picture_back_porch: ; Output the sync pulse of a picture line.
	bset TV_LATCH,#TV_SYNC

	; Are we about to trace a scanline?
	mov _vdc_active_line_counter,w0
	sub #240,w0
	bra c,picture_back_porch_done

	mov _vdc_frame_scan_ptr,w0
	mov #edsoffset(_vdc_scanline_buffer),w1

	push RCOUNT
	repeat #19
	mov [w0++],[w1++]
	pop RCOUNT

	mov w0,_vdc_frame_scan_ptr
	clr [w1]

picture_back_porch_done:
	
	mov #handle(picture_picture),w0
	mov _vdc_tv_line_offset_x,w1
	add #(8 * TV_TIMER_MS_PERIOD - 1),w1
	goto exit_retrace

picture_picture: ; Output the picture pulse of a picture line.

	mov _vdc_active_line_counter,w0
	sub #240,w0
	bra c,picture_picture_done

	; Switch DMA channels off then on again to reset it.

	bclr DMA1CON,#15 ; CHEN
	bset DMA1CON,#15 ; CHEN

	bclr DMA0CON,#15 ; CHEN
	bset DMA0CON,#15 ; CHEN

	; Switch SPI1 off to reset it.
	bclr SPI1STAT,#15 ; SPIEN

	; Reset timer 2.
	clr TMR2

	; Switch SPI1 on.	
	bset SPI1STAT,#15 ; SPIEN

picture_picture_done:

	inc _vdc_active_line_counter

	mov #(52 * TV_TIMER_MS_PERIOD - 1),w0
	mov _vdc_tv_line_offset_x,w1
	sub w0,w1,w1
	
	dec _vdc_tv_retrace_function_repeat_count
	bra z,picture_picture_to_short_sync
	; Repeat picture sync.
	mov #handle(picture_sync),w0
	goto exit_retrace
picture_picture_to_short_sync:
	; Switch to short sync.
	mov #6,w0
	mov w0,_vdc_tv_retrace_function_repeat_count
	mov #handle(short_sync_post_frame_sync),w0
	goto exit_retrace

short_sync_post_frame_sync: ; Output the sync portion of a short sync.
	bclr TV_LATCH,#TV_SYNC
	mov #handle(short_sync_post_frame_black),w0
	mov #(2 * TV_TIMER_MS_PERIOD - 1),w1
	goto exit_retrace

short_sync_post_frame_black: ; Output the sync portion of a short sync.
	bset TV_LATCH,#TV_SYNC
	mov #(30 * TV_TIMER_MS_PERIOD - 1),w1
	dec _vdc_tv_retrace_function_repeat_count
	bra z,short_sync_post_frame_to_long_sync
	; Repeat short sync.
	mov #handle(short_sync_post_frame_sync),w0
	goto exit_retrace
short_sync_post_frame_to_long_sync:
	
	; Check if the TV is still connected.
	mov RPOR2,w0
	and w0,#0b11111,w1
	mov w1,RPOR2

	bset TV_TRIS,#TV_PICTURE
	clr _vdc_tv_connected
	nop
	nop
	btss TV_PORT,#TV_PICTURE
	inc _vdc_tv_connected

	bclr TV_TRIS,#TV_PICTURE
	mov #0b0011100000000,w0
	ior w0,w1,w0
	mov w0,RPOR2

	; Switch to long sync.
	mov #5,w0
	mov w0,_vdc_tv_retrace_function_repeat_count
	mov #handle(long_sync_sync),w0
	goto exit_retrace

exit_retrace:

	mov w0,_vdc_retrace_function_ptr
	mov w1,PR1

	; Restore registers used by this function from the stack.
	pop w1
	pop w0
	
	; Clear the interrupt flag.
	bclr IFS0,#3 ; T1IF
	bset IEC0,#3 ; T1IE
	retfie

.end
